package ru.dbotthepony.mc.otm.container

import net.minecraft.core.HolderLookup
import net.minecraft.core.registries.BuiltInRegistries
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.NbtOps
import net.minecraft.resources.ResourceLocation
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.neoforged.neoforge.common.util.INBTSerializable
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.core.isNotEmpty
import ru.dbotthepony.mc.otm.core.nbt.set
import ru.dbotthepony.mc.otm.core.registryName
import ru.dbotthepony.mc.otm.data.getOrNull

open class ContainerSlot(
	final override val container: SlottedContainer,
	final override val slot: Int
) : ISlottedContainerSlot, INBTSerializable<CompoundTag> {
	private var _item: ItemStack = ItemStack.EMPTY

	final override var item: ItemStack
		get() = _item
		set(value) {
			_item = value
			setChanged()
		}

	private var observedItem = ItemStack.EMPTY

	// called from inside setChanged
	protected open fun notifyChanged(old: ItemStack) {}

	final override fun setChanged() {
		observeChanges()
	}

	/**
	 * Called when slot needs to be cleared of any data present
	 */
	open fun clear() {
		_item = ItemStack.EMPTY

		if (observedItem.isNotEmpty) {
			notifyChanged(observedItem)
		}

		observedItem = ItemStack.EMPTY
	}

	fun observeChanges(): Boolean {
		if (observedItem.count != item.count || !ItemStack.isSameItemSameComponents(item, observedItem)) {
			notifyChanged(observedItem)
			observedItem = item.copy()
			container.notifyChanged()
			return true
		}

		return false
	}

	override fun remove(): ItemStack {
		val item = item

		if (item.isEmpty) {
			return ItemStack.EMPTY
		} else {
			this.item = ItemStack.EMPTY
			return item
		}
	}

	override val maxStackSize: Int
		get() = Item.DEFAULT_MAX_STACK_SIZE

	override fun remove(count: Int): ItemStack {
		val item = item

		if (item.isEmpty) {
			return ItemStack.EMPTY
		}

		if (item.count >= count) {
			this.item = ItemStack.EMPTY
			return item
		} else {
			val split = item.split(count)
			setChanged()
			return split
		}
	}

	override fun serializeNBT(provider: HolderLookup.Provider): CompoundTag {
		return CompoundTag().also {
			it["item"] = ItemStack.OPTIONAL_CODEC.encodeStart(provider.createSerializationContext(NbtOps.INSTANCE), item)
				.getOrThrow { RuntimeException("Unable to serialize $item in slot $slot: $it") }
		}
	}

	override fun deserializeNBT(provider: HolderLookup.Provider, nbt: CompoundTag) {
		_item = ItemStack.OPTIONAL_CODEC.decode(provider.createSerializationContext(NbtOps.INSTANCE), nbt["item"])
			.ifError { LOGGER.error("Unable to deserialize item at slot $slot: ${it.message()}") }
			.getOrNull()?.first ?: ItemStack.EMPTY

		observedItem = item.copy()
		notifyChanged(ItemStack.EMPTY)
	}

	open class Simple(
		protected val listener: (new: ItemStack, old: ItemStack) -> Unit,
		protected val maxStackSize: Int = Item.DEFAULT_MAX_STACK_SIZE,
	) : SlottedContainerBuilder.SlotProvider {
		protected open inner class Instance(container: SlottedContainer, slot: Int) : ContainerSlot(container, slot) {
			override val maxStackSize: Int
				get() = this@Simple.maxStackSize

			override fun notifyChanged(old: ItemStack) {
				super.notifyChanged(old)
				listener(item, old)
			}
		}

		override fun create(container: SlottedContainer, index: Int): ContainerSlot {
			return Instance(container, index)
		}
	}

	open class Filtered(
		listener: (new: ItemStack, old: ItemStack) -> Unit,
		maxStackSize: Int = Item.DEFAULT_MAX_STACK_SIZE,
	) : Simple(listener, maxStackSize) {
		protected open inner class Instance(container: SlottedContainer, slot: Int) : Simple.Instance(container, slot), IFilteredSlottedContainerSlot {
			override var filter: Item? = null
				set(value) {
					if (field !== value) {
						field = value
						container.notifyChanged()
					}
				}

			override fun clear() {
				super.clear()
				filter = null
			}

			override fun serializeNBT(provider: HolderLookup.Provider): CompoundTag {
				return super.serializeNBT(provider).also {
					if (filter != null)
						it["filter"] = filter!!.registryName!!.toString()
				}
			}

			override fun deserializeNBT(provider: HolderLookup.Provider, nbt: CompoundTag) {
				super.deserializeNBT(provider, nbt)

				if ("filter" in nbt) {
					filter = BuiltInRegistries.ITEM.get(ResourceLocation.parse(nbt.getString("filter")))
				}
			}
		}

		override fun create(container: SlottedContainer, index: Int): ContainerSlot {
			return Instance(container, index)
		}
	}

	companion object {
		private val LOGGER = LogManager.getLogger()
	}
}
