package ru.dbotthepony.mc.otm.container

import com.mojang.serialization.Codec
import com.mojang.serialization.codecs.RecordCodecBuilder
import net.minecraft.core.HolderLookup
import net.minecraft.core.registries.BuiltInRegistries
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.ListTag
import net.minecraft.nbt.NbtOps
import net.minecraft.nbt.Tag
import net.minecraft.world.entity.player.Player
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.neoforged.neoforge.common.util.INBTSerializable
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.data.codec.minRange
import java.util.function.Predicate

class SlottedContainer(
	slots: Collection<SlottedContainerBuilder.SlotProvider>,
	private val stillValid: Predicate<Player>,
	private val globalChangeListeners: Array<Runnable>
) : ISlottedContainer, INBTSerializable<Tag> {
	private val slots: Array<ContainerSlot>

	init {
		val itr = slots.iterator()
		this.slots = Array(slots.size) { itr.next().create(this, it) }
	}

	override val hasFilterableSlots: Boolean = this.slots.any { it is IFilteredContainerSlot }
	private var suppressListeners = false

	override fun clearContent() {
		suppressListeners = true

		try {
			slots.forEach { it.remove() }
			notifyChanged()
		} finally {
			suppressListeners = false
		}
	}

	override fun containerSlot(slot: Int): ISlottedContainerSlot {
		return slots[slot]
	}

	fun notifyChanged() {
		if (suppressListeners) return
		globalChangeListeners.forEach { it.run() }
	}

	// called by outside code (vanilla and other unaware mods)
	override fun setChanged() {
		suppressListeners = true
		var hasChanges = false

		try {
			slots.forEach { hasChanges = it.observeChanges() || hasChanges }
		} finally {
			suppressListeners = false

			if (hasChanges)
				notifyChanged()
		}
	}

	override fun getContainerSize(): Int {
		return slots.size
	}

	override fun stillValid(player: Player): Boolean {
		return stillValid.test(player)
	}

	private data class LegacySerializedItem(val item: ItemStack, val slot: Int) {
		companion object {
			val CODEC: Codec<LegacySerializedItem> = RecordCodecBuilder.create {
				it.group(
					ItemStack.OPTIONAL_CODEC.fieldOf("item").forGetter { it.item },
					Codec.INT.minRange(0).fieldOf("slot").forGetter { it.slot },
				).apply(it, ::LegacySerializedItem)
			}
		}
	}

	private data class LegacySerializedFilter(val item: Item, val slot: Int)  {
		companion object {
			val CODEC: Codec<LegacySerializedFilter> = RecordCodecBuilder.create {
				it.group(
					BuiltInRegistries.ITEM.byNameCodec().fieldOf("item").forGetter { it.item },
					Codec.INT.minRange(0).fieldOf("slot").forGetter { it.slot },
				).apply(it, ::LegacySerializedFilter)
			}
		}
	}

	private data class LegacySerializedState(
		val items: List<LegacySerializedItem>,
		val filters: List<LegacySerializedFilter>
	) {
		companion object {
			val CODEC: Codec<LegacySerializedState> = RecordCodecBuilder.create {
				it.group(
					Codec.list(LegacySerializedItem.CODEC).fieldOf("items").forGetter { it.items },
					Codec.list(LegacySerializedFilter.CODEC).fieldOf("filters").forGetter { it.filters },
				).apply(it, ::LegacySerializedState)
			}
		}
	}

	private val lostItems = ArrayList<CompoundTag>()

	override fun serializeNBT(provider: HolderLookup.Provider): ListTag {
		return ListTag().also {
			for (slot in slots) {
				it.add(slot.serializeNBT(provider))
			}

			it.addAll(lostItems)
		}
	}

	override fun deserializeNBT(provider: HolderLookup.Provider, nbt: Tag) {
		lostItems.clear()
		slots.forEach { it.clear() }

		if (nbt is CompoundTag) {
			// legacy container
			LegacySerializedState.CODEC.decode(provider.createSerializationContext(NbtOps.INSTANCE), nbt)
				.resultOrPartial { LOGGER.error("Error deserializing container: $it") }
				.ifPresent {
					val (items, filters) = it.first

					// excessive items will be lost
					for ((item, slot) in items) {
						if (slot in 0 until containerSize) {
							slots[slot].item = item
						}
					}

					for ((filter, slot) in filters) {
						if (slot in 0 until containerSize) {
							val getSlot = slots[slot]

							if (getSlot is IFilteredContainerSlot) {
								getSlot.filter = filter
							}
						}
					}
				}
		} else if (nbt is ListTag) {
			// normal container
			for ((i, element) in nbt.withIndex()) {
				if (element !is CompoundTag) {
					LOGGER.error("Deserializing mattery container: Expected compound tag at $i, got $element")
					continue
				}

				if (i in 0 until containerSize) {
					slots[i].deserializeNBT(provider, element)
				} else {
					lostItems.add(element)
				}
			}
		} else {
			LOGGER.error("Unable to deserialize mattery container, expected CompoundTag or ListTag, got $nbt")
		}

		notifyChanged()
	}

	companion object {
		private val LOGGER = LogManager.getLogger()
	}
}
